// Cellule class 3D
// Creation: 25/11/2008
// Last modification: 20/01/2009


import java.util.Vector;
import java.lang.Math.*;


/** 
 * Class creating a cell
 */

public class Cell {


    private int occup;  
    private Point P;
    private Vector Neighbours;
    private Vector Probas; 

    private double arrival;
    private Vector voronoi;

    private String typeGrid;

    /**
     * Constructor of a cell with its coordinates (x,y) and without any neighbour
     * @param Q Cell point
     */

    public Cell (Point Q, String grid) {
	P = Q;
	Neighbours = new Vector();
	voronoi = new Vector();
	Probas = new Vector();
	occup = -1;
	typeGrid = grid;
    }



    /** 
     * Returns the occupation of the cell
     * @return State of the cell: -1 if empty, else the number of the particle
     */
    public int getOccupied() {
	return occup;
    }



    /** 
     * Returns the cell point
     * @return Cell point
     */
    public Point getPoint() {
	return P;
    }



    /** 
     * Returns the cartesian coordinates of the cell
     * @return Vector giving the cell position
     */
    public Vector getCartCoordinates() {
	return P.getCartCoordinates();
    }



    /** 
     * Returns the spherical coordinates of the cell
     * @return Vector giving the cell position
     */
    public Vector getSphereCoordinates() {
	return P.getSphereCoordinates();
    }



    /** 
     * Returns the neighbour list of the cell
     * @return Neighbour list of the cell
     */
    public Vector getNeighbours() {
	return Neighbours;
    }


    /** 
     * Returns the probability list linked to the neighbours
     * @return probability list of the cell
     */
    public Vector getProbas() {
	return Probas;
    }


    
    /** 
     * Returns the number of neighbours of the cell
     * @return Number of neighbours of the cell
     */
    public int getNumberNeighbours() {
	return Neighbours.size();
    }


    
    /** 
     * Returns the probability that a particle arrives on the cell
     * @return Probability that a particle arrives on the cell
     */
    public double getArrival() {
	return arrival;
    }



    /** 
     * Displays the position of the cell
     */
    public void printPosition() {
	System.out.println("Cartesiennes      x= " + this.getCartCoordinates().elementAt(0) +  " y= " + this.getCartCoordinates().elementAt(1) + " z= " + this.getCartCoordinates().elementAt(2) + "\n" +  "Spheriques        r= " + this.getSphereCoordinates().elementAt(0) +  " theta= " + this.getSphereCoordinates().elementAt(1) + " phi= " + this.getSphereCoordinates().elementAt(2) + "\n");
    }
    


    /** 
     * Displays the essential information of the cell
     */
    public void printInfo() {
	System.out.println("Infos de la cellule: ");
	this.printPosition();
	System.out.println("Etat d'occupation: "  + this.getOccupied() );
	System.out.println("Nombre de voisins: "  + this.getNumberNeighbours() ); 
	System.out.println("Proba d'arrive : "  + this.getArrival() );	
    }



    /** 
     * Displays the neighbours of the cell, and the associated probabilities
     */
    public void printNeighbours() {
	int n = this.getNumberNeighbours();
	Vector lv = this.getNeighbours();
	Vector lp = this.getProbas();
	for (int i=0; i<n; i++) {
	    Double q1v = (Double)((Cell)(lv.elementAt(i))).getCartCoordinates().elementAt(0);

	    Double q2v = (Double)((Cell)(lv.elementAt(i))).getCartCoordinates().elementAt(1);

	    Double q3v = (Double)((Cell)(lv.elementAt(i))).getCartCoordinates().elementAt(2);

	    System.out.println("Voisin " + i + " x= " + q1v + " y= " + q2v + " z= " + q3v 
			       + " proba= " + lp.elementAt(i));	    
		}
    }



    /** 
     * Displays the neighbours relatively to the position of the cell
     */
    public void printNeighboursRelat() {
	int n = this.getNumberNeighbours();
	Vector lv = this.getNeighbours();
	Vector lp = this.getProbas();
	for (int i=0; i<n; i++) {
	    double q1v = ((Double)((Cell)(lv.elementAt(i))).getCartCoordinates().elementAt(0)).doubleValue()
		- ((Double)(this.getCartCoordinates().elementAt(0))).doubleValue();

	    double q2v = ((Double)((Cell)(lv.elementAt(i))).getCartCoordinates().elementAt(1)).doubleValue()
		-((Double)(this.getCartCoordinates().elementAt(1))).doubleValue() ;			   

	    double q3v = ((Double)((Cell)(lv.elementAt(i))).getCartCoordinates().elementAt(2)).doubleValue()
		-((Double)(this.getCartCoordinates().elementAt(2))).doubleValue();

	    System.out.println("Voisin " + i + " x= " + q1v + " y= " + q2v + " z= " + q3v 
			       + " proba= " + lp.elementAt(i));	    
		}
    }



   /** 
     * Displays the vertices of the Voronoi cell
     */
    public void printVoronoi() {

	Vector lv = this.getVoronoi();
	for (int i=0; i<lv.size(); i++) {
	    Double q1v = (Double)((Point)(lv.elementAt(i))).getCartCoordinates().elementAt(0);

	    Double q2v = (Double)((Point)(lv.elementAt(i))).getCartCoordinates().elementAt(1);

	    Double q3v = (Double)((Point)(lv.elementAt(i))).getCartCoordinates().elementAt(2);

	    System.out.println("Sommet " + i + " x= " + q1v + " y= " + q2v + " z= " + q3v );	    
		}
    }


    
    /** 
     * Checks if a cell is a neighbour of the considered cell
     * @param cell Test cell
     */
    public boolean checkNeighbour (Cell cell) {
	if (Neighbours.contains(cell)) {
	    return true;
	}
	else {
	    return false;
	}
    }



    /** 
     * Adds a neighbour to the cell
     * @param cell Cell to add
     * @param p Associated probability
     */
    public void addNeighbour (Cell cell, double p) {
	if (this.checkNeighbour(cell) == false) {
	    Neighbours.addElement(cell);
	    Double ref = new Double(p);
	    Probas.addElement(ref);
	}
    }


    
    /** 
     * Removes a neighbour to the cell 
     * @param cell Cell to remove
     */
    public void removeNeighbour (Cell cell) {
	if (this.checkNeighbour(cell) == true) {
	    int i = Neighbours.indexOf(cell);
	    Neighbours.removeElementAt(i);
	    Probas.removeElementAt(i);
	}
    }
    


    /** 
     * Returns the euclidian distance between 2 cells
     * @param cell Cell 
     * @return Distance between the 2 cells
     */
    public double eucliDist(Cell cell) {
	Point Q = cell.getPoint();
	return P.eucliDist(Q);
    }



    /** 
     * Changes the occupation of the cell
     * @param n New value of the occupation state
     */
    public void setOccupied (int n) {
	this.occup = n;
    }



    /**
     * Generates the cell list of a spherical grid from a location structure
     * @param grid the concerned spherical grid
     * @param graph the location structure used to build the cell list
     * @return cellList Vector containing all the generated cells 
     */
    public static Vector cellGeneration(SphereGrid grid, LocationStructure graph){

	Cell root = grid.getRoot();

	Vector cellList = new Vector();
	Vector pointList = new Vector();

	Vector leafList = graph.getLeafList();

	for(int i=0;i<leafList.size();i++){
	        
	    Triangle tr = ((GraphNode)(leafList.elementAt(i))).getTriangle();

	    Point P1 = tr.getS1();
	    Point P2 = tr.getS2();
	    Point P3 = tr.getS3();

	    Cell cell1 = null;
	    Cell cell2 = null;
	    Cell cell3 = null;

	    int test = 0;

	    for (int j=0; j <cellList.size();j++){

		cell1 = (Cell)cellList.elementAt(j);
		if (P1 == cell1.getPoint()){
		    test =1;
		    break;
		}
	    }
	    if (test == 0){
		cell1 = new Cell(P1,"sphere");
		cellList.addElement(cell1);
	    }

	    test=0;
	    
	    for (int j=0; j <cellList.size();j++){

		cell2 = (Cell)cellList.elementAt(j);
		if (P2 == cell2.getPoint()){
		    test =1;
		    break;
		}
	    }
	    if (test == 0){
		cell2 = new Cell(P2,"sphere");
		cellList.addElement(cell2);
	    }

	    test=0;
	    
	    for (int j=0; j <cellList.size();j++){

		cell3 = (Cell)cellList.elementAt(j);
		if (P3 == cell3.getPoint()){
		    test =1;
		    break;
		}
	    }
	    if (test == 0){
		cell3 = new Cell(P3,"sphere");
		cellList.addElement(cell3);
	    }

	    
	    cell1.addNeighbour(cell2,0.0);
	    cell1.addNeighbour(cell3,0.0);

	    cell2.addNeighbour(cell1,0.0);
	    cell2.addNeighbour(cell3,0.0);

	    cell3.addNeighbour(cell1,0.0);
	    cell3.addNeighbour(cell2,0.0);

	}

	double surfTot = 0.0; 

	for(int i=0;i<cellList.size();i++){
	    Cell c = (Cell)cellList.elementAt(i);

	    Vector neigh = c.getNeighbours();
	    double sum = 0.0;
	    Vector prob = new Vector();

	    for(int j=0; j<neigh.size(); j++){
		sum = sum +(1.0 / c.eucliDist((Cell)neigh.elementAt(j)));
	    }
	    for(int j=0; j<neigh.size(); j++){
		prob.addElement(new Double((1.0 / c.eucliDist((Cell)neigh.elementAt(j)))/sum));
	    }
	    
	    c.setProbas(prob);

	    // The neighbours are spacially sorted
	    // calculation of the voronoi vertices
	    
	    c.generateVoronoi();
	    surfTot = surfTot + c.getArrival();
	}
	
	// normalization of the arrival probabilities

	for(int i=0;i<cellList.size();i++){
	
	    Cell c = (Cell)cellList.elementAt(i);
	    double probRoot = c.getArrival()/surfTot;
	    root.addNeighbour(c,probRoot);
	    c.setArrival(probRoot);
	}
	
	return cellList;

    }



    /** 
     * Generates the Voronoi vertices of the Cell
     */
    public void generateVoronoi() {
	
	if(typeGrid.equals("sphere")){
		// The neighbours are spacially sorted
		this.sortNeighbours();
	}

	int n = this.getNumberNeighbours();
	Vector N = this.getNeighbours();
	
	Vector vorList = new Vector(); 

	double proba = 0.0;
	Point O = this.getPoint(); 
	
	// Les voisins doivent etre classes !

	for (int i=0; i<n; i++) {

	    Cell C1;
	    Cell C2;
	    Cell C3;

	    if (i == n-1) {
		C1 = (Cell)((Vector)N).elementAt(i);
		C2 = (Cell)((Vector)N).elementAt(0);
		C3 = (Cell)((Vector)N).elementAt(1);
	    }
	    else if (i == n-2) {	
		C1 = (Cell)((Vector)N).elementAt(i);
		C2 = (Cell)((Vector)N).elementAt(i+1);
		C3 = (Cell)((Vector)N).elementAt(0);
	    }
	    else {
		C1 = (Cell)((Vector)N).elementAt(i);
		C2 = (Cell)((Vector)N).elementAt(i+1);
		C3 = (Cell)((Vector)N).elementAt(i+2);
	    }
	    
	    Point A = C1.getPoint();
	    Point B = C2.getPoint();
	    Point C = C3.getPoint();

	    Point I = O.intersec(A,B);
	    Point J = O.intersec(B,C);
	    
	    vorList.addElement(I);

	    Triangle T = new Triangle(O,I,J);
	    double rec = T.area();
	    
	    proba = proba + rec;
	}
    
	this.arrival = proba;
	
	this.voronoi =  vorList;

    }
    


    /**
    * Setter for the "arrival" attribute
    * @param value set value
    */
    public void setArrival(double value){
	arrival = value;
    }



    /**
    * Sorts geometrically all the neighbours of the cell if you consider a rotation around the cell.
    *
    */
    private void sortNeighbours(){

	Vector up = new Vector();
	Vector upScal = new Vector();

	Vector down = new Vector();
	Vector downScal = new Vector();

	Vector tempList = new Vector();

	Cell cellRef = (Cell)Neighbours.elementAt(0);



	Point ori = new Point(0,0,0,0);
	Point pRef = cellRef.getPoint();
	    
	Vect ori_P = new Vect(ori,P);
	Vect normal  = ori_P.propVect(ori_P.norm());

	Vect p_pRef = new Vect(P, pRef);


	double compoFact1 = normal.scalar(p_pRef);
	Vect compo1 = normal.propVect(-compoFact1);
	Vect p_pRefP = p_pRef.sum(compo1);

	// separation des voisins en deux groupes prliminaires
	// grace au produit vectoriel
	// attention : on ne traite pas le premier point de la liste
	// puisqu'il est pris comme reference

	for(int i=1; i< Neighbours.size(); i++){
	    
	    Cell neigh = (Cell)Neighbours.elementAt(i);
	    Point np = neigh.getPoint();
	    
	    Vect p_np = new Vect(P,np);

	    double compoFact2 = normal.scalar(p_np);
	    Vect compo2 = normal.propVect(-compoFact2);
	    Vect p_npP = p_np.sum(compo2); 
	    
	    if ( ((p_pRefP.cross(p_npP)).scalar(ori_P))>=0 ){
		up.addElement(neigh);
	    }
	    else{
		down.addElement(neigh);
	    }    
	}

	// sort the neighbours of the "up" list
	// from the one maximising the scalar product
	// to the one minimising the scalar product

	while(up.size()>0){
	    double maxScal = -1000.0;
	    int index = -1;
	    for (int i=0; i<up.size(); i++){

		Cell neigh = (Cell)up.elementAt(i);
		Point np = neigh.getPoint();

		Vect p_np = new Vect(P,np);

		double compoFact2 = normal.scalar(p_np);
	        Vect compo2 = normal.propVect(-compoFact2);
	        Vect p_npP = p_np.sum(compo2);

		double tempScal = p_npP.scalar(p_pRefP)/(p_npP.norm()*p_pRefP.norm());
		if (tempScal>maxScal){
		    maxScal=tempScal;
		    index=i;
		}
	    
	    }
	    upScal.addElement(up.elementAt(index));
	    up.removeElementAt(index);
	}


	// sort the neighbours of the "down" list
	// from the one minimising the scalar product
	// to the one maximising the scalar product

	while(down.size()>0){
	    double minScal = 1000.0;
	    int index = -1;
	    for (int i=0; i<down.size(); i++){

		Cell neigh = (Cell)down.elementAt(i);
		Point np = neigh.getPoint();

		Vect p_np = new Vect(P,np);

		double compoFact2 = normal.scalar(p_np);
	        Vect compo2 = normal.propVect(-compoFact2);
	        Vect p_npP = p_np.sum(compo2);

		double tempScal = p_npP.scalar(p_pRefP)/(p_npP.norm()*p_pRefP.norm());
		if (tempScal<minScal){
		    minScal=tempScal;
		    index=i;
		}
	    
	    }
	   downScal.addElement(down.elementAt(index));
	   down.removeElementAt(index);
	}


	// build the new neighbour list

	tempList.addElement(cellRef);
	for(int i=0; i<upScal.size(); i++){
	    tempList.addElement(upScal.elementAt(i));
	}
	for(int i=0; i<downScal.size(); i++){
	    tempList.addElement(downScal.elementAt(i));
	}

	Neighbours = tempList;

    }



    /** 
     * Setter for the probability list
     * @param p Vector containing Double objects which represent the desired associated probabilities
     */
    public void setProbas(Vector p) {
	Probas = p;
    }


    /**
    * Getter for the "voronoi" attribute
    */
    public Vector getVoronoi(){
	return voronoi;
    }

}
    
