// Triangle class
// Creation : 21/11/2008
// Last modification : 20/01/2009


import java.lang.Math.*;
import java.util.Vector;


/** 
 * Class representing a triangle
 */
public class Triangle {

    private Point S1;
    private Point S2;
    private Point S3;

    private Vector vertexList;


    /**
     * Constructor of a triangle definied by 3 sommets
     * @param P Sommet #1 
     * @param Q Sommet #2
     * @param R Sommet #3
     */
    public Triangle (Point P, Point Q, Point R) {
	S1 = P;
	S2 = Q;
	S3 = R;

	vertexList = new Vector();
	vertexList.addElement(S1);
	vertexList.addElement(S2);
	vertexList.addElement(S3);
    }



    /** 
     * Returns the #1 peak of the triangle
     * @return S1
     */
    public Point getS1() {
	return S1;
    }



    /** 
     * Returns the #2 peak of the triangle
     * @return S2
     */
    public Point getS2() {
	return S2;
    }



    /** 
     * Returns the #3 peak of the triangle
     * @return S3
     */
    public Point getS3() {
	return S3;
    }



    /** 
     * Returns the vextex list of the triangle
     * @return Vertex list
     */

    public Vector getVertex(){
	return vertexList;
    }
    


    /** 
     * Returns the vector OG where G is the gravity centre of the triangle
     * @return Vector 0G
     */
    public Vect gravity() {
	Point S1 = this.getS1();
	Point S2 = this.getS2();
	Point S3 = this.getS3();
	Point O = new Point(0,0,0,0);

	Vect A = new Vect(O,S1);
	Vect B = new Vect(O,S2);
	Vect C = new Vect(O,S3);

	Vect R = A.sum(B);
	R = R.sum(C);
	R = R.propVect(1./3.);
	return R;
    }
    


    /** 
     * Displays the cartesian coordinates of the triangle peaks
     */
    public void printCoordinates() {
	Point S1 = this.getS1();
	Point S2 = this.getS2();
	Point S3 = this.getS3();
	Point O = new Point(0,0,0,0);

	Vect A = new Vect(O,S1);
	Vect B = new Vect(O,S2);
	Vect C = new Vect(O,S3);
	
	System.out.println("S1:-" + S1.getName() +   "->    x= " + A.getCoordinates().elementAt(0) +  " y= " + A.getCoordinates().elementAt(1) + " z= " + A.getCoordinates().elementAt(2) + "\n");
	System.out.println("S2:-" + S2.getName() +   "->    x= " + B.getCoordinates().elementAt(0) +  " y= " + B.getCoordinates().elementAt(1) + " z= " + B.getCoordinates().elementAt(2) + "\n");
	System.out.println("S3:-" + S3.getName() +  "->    x= " + C.getCoordinates().elementAt(0) +  " y= " + C.getCoordinates().elementAt(1) + " z= " + C.getCoordinates().elementAt(2) + "\n");
    }
    


    /** 
     * Returns the coefficients of the equation of the plan containing the triangle:  Ax + By + Cz + D = 0
     * @return Coefficients vector
     */
    public Vector plan() {
	Point S1 = this.getS1();
	Point S2 = this.getS2();
	Point S3 = this.getS3();

	Vect A = new Vect(S1,S2);
	Vect B = new Vect(S1,S3);

	Vect N = A.cross(B);

	N = N.normalize();

	Point O = new Point(0,0,0,0);
	Vect C = new Vect(O,S1);
	
	double D = -N.scalar(C);

	Vector ABC = N.getCoordinates();
	Vector result = new Vector(4);
	
	result.addElement(ABC.elementAt(0));
	result.addElement(ABC.elementAt(1));
	result.addElement(ABC.elementAt(2));
	Double DD = new Double(D);
	result.addElement(DD);
	
	return result;
    }
    


    /** 
     * Returns the projection of a point on the plan containing the triangle
     * @param P Point we want to project
     * @return Projected point
     */
    public Point proj(Point P) {

	Vector Q = P.getCartCoordinates();
	double r = ((Double)((Vector)Q).elementAt(0)).doubleValue();
	double s = ((Double)((Vector)Q).elementAt(1)).doubleValue();
	double t = ((Double)((Vector)Q).elementAt(2)).doubleValue();

	Vector V = this.plan();
	double a = ((Double)((Vector)V).elementAt(0)).doubleValue();
	double b = ((Double)((Vector)V).elementAt(1)).doubleValue();
	double c = ((Double)((Vector)V).elementAt(2)).doubleValue();
	double d = ((Double)((Vector)V).elementAt(3)).doubleValue();
	
	Point O = new Point(0,0,0,0);
	Point U = new Point(a,b,c,0);
	Vect N = new Vect(O,U);

	N = N.normalize();

	double x = (b*b + c*c)*r - a*b*s - a*c*t - d*a;
	double y = -a*b*r + (a*a+c*c)*s - b*c*t - d*b;
	double z = -a*c*r - b*c*s + (a*a+b*b)*t - d*c;

	Point result = new Point(x,y,z,0);

	return result;
    }
    


    /** 
     * Returns the area of the triangle
     * @return Triangle area
     */
    public double area() {

	Point S1 = this.getS1();
	Point S2 = this.getS2();
	Point S3 = this.getS3();
		
	Vect S1S2 = new Vect(S1,S2);
	Vect S1S3 = new Vect(S1,S3);

	double area = 1./2. * (S1S2.cross(S1S3)).norm(); 
	return area;
    }    
    


    /**
     * Checks if the triangle owns the two points specified as arguments
     * NB: comparison of the adresses not of the coordinates themselves
     * @param p1 one of the two points to be checked
     * @param p2 one of the two points to be checked
     * @return answer 
     */
    public boolean owns(Point p1, Point p2){

	int count = 0;
	if((S1==p1) || (S2==p1) || (S3==p1)){
	    count++;
	}
	if((S1==p2) || (S2==p2) || (S3==p2)){
	    count++;
	}
	
	if(count == 2){
	    return true;
	}
	else{
	    return false;
	}
    }



    /**
     * Checks if the point specified as argument is contained in the triangle
     * @param P the point to be checked
     * @return answer
     */
    public boolean contains(Point P){

	Point Ori = new Point(0,0,0,0);

	Vect P1= new Vect(Ori, S1);
	Vect P2= new Vect(Ori, S2);
	Vect P3= new Vect(Ori, S3);

	Vect G = this.gravity();
	Vect VP = new Vect(Ori, P);

	Vect N1 = P1.cross(P2);
	Vect N2 = P2.cross(P3);
	Vect N3 = P3.cross(P1);
	
	if ( ((N1.scalar(VP))*(N1.scalar(G)) >= 0.0) && 
	     ((N2.scalar(VP))*(N2.scalar(G)) >= 0.0) && 
	     ((N3.scalar(VP))*(N3.scalar(G)) >= 0.0) ){
	    return true;
	}
	else{
	    return false;
	}

    }
   


    /** 
     * Return the coordinates of the circumcenter of triangle and the radius of the circumcircle
     * @return Vector containing the coordinates and the radius
     */
    public Vector circum() {

	Vector result  = new Vector(4);

	Point A = this.getS1();
	Point B = this.getS2();
	Point C = this.getS3();
	Point O = new Point(0,0,0,0);

	Vect P1 = new Vect(O,A);
	Vect P2 = new Vect(O,B);
	Vect P3 = new Vect(O,C);

	Vect P12 = P2.diff(P1);
	Vect P23 = P3.diff(P2);
	Vect P31 = P1.diff(P3);

	double r = P12.norm() * P23.norm() * P31.norm() ;
	double d = 2.0 * (P12.cross(P23)).norm(); 
	r = r / d;

	Double radius = new Double(r);

	double alpha = P23.norm() * P23.norm() * P12.scalar(P31.propVect(-1.0));
	double beta  = P31.norm() * P31.norm() * P23.scalar(P12.propVect(-1.0));
	double gamma = P12.norm() * P12.norm() * P31.scalar(P23.propVect(-1.0));
	
	alpha = alpha / (d*d) * 2.0;
	beta = beta / (d*d) *2.0;
	gamma = gamma / (d*d) * 2.0;


	Vect U1 = P1.propVect(alpha);
	Vect U2 = P2.propVect(beta);
	Vect U3 = P3.propVect(gamma);

	Vect PC = (U1.sum(U2)).sum(U3);	

	Vector coord = PC.getCoordinates();
	
	result.addElement(coord.elementAt(0));
	result.addElement(coord.elementAt(1));
	result.addElement(coord.elementAt(2));
	result.addElement(radius);

	return result;

    }



    /** 
     * Return the boolean verifying if the circumcircle of the triangle contains the point P
     * @param P the Point to be check
     * @return Boolean
     */
    public boolean circumContains(Point P) {
	Vector C = this.circum();

	Double x = (Double)((Vector)C).elementAt(0);
	Double y = (Double)((Vector)C).elementAt(1);
	Double z = (Double)((Vector)C).elementAt(2);

	double xc = x.doubleValue();
	double yc = y.doubleValue();
	double zc = z.doubleValue();

	Point O = new Point(xc,yc,zc,0);
	Vect OP = new Vect(O,P);
	
	Double r = (Double)((Vector)C).elementAt(3);
	double radius = r.doubleValue();

	double l = OP.norm();
	

	if (((l-radius)/l) < 1E-10) {
	    return true;
	}
	else {
	    return false;
	}
	
    }
    


}
